@@ -26,23 +26,27 @@ import json
 
 from pathlib import Path
 from threading import Thread
-from socketIO_client import SocketIO, SocketIONamespace
+# from socketIO_client import SocketIO, SocketIONamespace
+from socketio import (
+    ClientNamespace,
+    Client as SocketIO
+)
 from gevent.pywsgi import WSGIServer
 
 from . import globals as cs
 
+from vantage6.common.docker_addons import ContainerKillListener
 from vantage6.node.docker_manager import DockerManager
 from vantage6.node.server_io import NodeClient
 from vantage6.node.proxy_server import app
 from vantage6.node.util import logger_name
 
 
-class NodeTaskNamespace(SocketIONamespace):
+class NodeTaskNamespace(ClientNamespace):
     """Class that handles incoming websocket events."""
 
     # reference to the node objects, so a callback can edit the
     # node instance.
-    # FIXME: why is this a *class* attribute?
     node_worker_ref = None
 
     def __init__(self, *args, **kwargs):
@@ -50,17 +54,6 @@ class NodeTaskNamespace(SocketIONamespace):
         """
         super().__init__(*args, **kwargs)
         self.log = logging.getLogger(logger_name(__name__))
-        self.node_worker_ref = None
-
-    def set_node_worker(self, node_worker):
-        """ Reference Node that created this Namespace.
-
-            This way we can call methods from the nodeworking, allowing
-            for actions to be taken.
-
-            :param node_worker: Node object
-        """
-        self.node_worker_ref = node_worker
 
     def on_message(self, data):
         self.log.info(data)
@@ -210,23 +203,41 @@ class Node(object):
             self.ctx.config.get("docker_registries", [])
         )
 
+        databases = {}
+
+        # Check that the `default` database label is present. If this is
+        # not the case, older algorithms will break
+        db_labels = self.config['databases'].keys()
+        if 'default' not in db_labels:
+            self.log.error("'default' database not specified in the config!")
+            self.log.debug(f'databases in config={db_labels}')
+
         # If we're running in a docker container, database_uri would point
         # to a path on the *host* (since it's been read from the config
         # file). That's no good here. Therefore, we expect the CLI to set
-        # the environment variable for us. This has the added bonus that we
+        # the environment variables for us. This has the added bonus that we
         # can override the URI from the command line as well.
-        default_uri = self.config['databases']['default']
-        database_uri = os.environ.get('DATABASE_URI', default_uri)
+        for label in db_labels:
+            label_upper = label.upper()
+            if self.__docker.running_in_docker():
+                uri = os.environ[f'{label_upper}_DATABASE_URI']
+            else:
+                uri = self.config['databases'][label]
 
-        if Path(database_uri).exists():
-            # We'll copy the file to the folder `data` in our task_dir.
-            self.log.info(f'Copying {database_uri} to {task_dir}')
-            shutil.copy(database_uri, task_dir)
+            db_is_file = Path(uri).exists()
+            if db_is_file:
+                # We'll copy the file to the folder `data` in our task_dir.
+                self.log.info(f'Copying {uri} to {task_dir}')
+                shutil.copy(uri, task_dir)
 
-            # Since we've copied the database to the folder 'data' in the root
-            # of the volume: '/data/<database.csv>'. We'll just keep the
-            # basename (i.e. filename + ext).
-            database_uri = os.path.basename(database_uri)
+                # Since we've copied the database to the folder 'data' in the
+                # root of the volume: '/data/<database.csv>'. We'll just keep
+                # the basename (i.e. filename + ext).
+                uri = os.path.basename(uri)
+
+                self.__docker.database_is_file = True
+
+            databases[label] = {'uri': uri, 'is_file': db_is_file}
 
         # Connect to the isolated algorithm network *only* if we're running in
         # a docker container.
@@ -237,7 +248,11 @@ class Node(object):
             )
 
         # Let's keep it safe
-        self.__docker.set_database_uri(database_uri)
+        self.__docker.set_databases(databases)
+
+        # Load additional environment vars for the algorithms. This is
+        # for example usefull when a password is needed for the database
+        self.__docker.algorithm_env = self.config.get('algorithm_env', {})
 
         # Thread for sending results to the server when they come available.
         self.log.debug("Start thread for sending messages (results)")
@@ -356,12 +371,14 @@ class Node(object):
 
         # Run the container. This adds the created container/task to the list
         # __docker.active_tasks
+        local_db = task['database'] if task['database'] else 'default'
         self.__docker.run(
             result_id=taskresult["id"],
             image=task["image"],
             docker_input=taskresult['input'],
             tmp_vol_name=vol_name,
-            token=token
+            token=token,
+            database=local_db
         )
 
     def __listening_worker(self):
@@ -396,12 +413,13 @@ class Node(object):
 
             # notify all of a crashed container
             if results.status_code:
-                self.socket_tasks.emit(
+                self.socketIO.emit(
                     'container_failed',
                     self.server_io.id,
                     results.status_code,
                     results.result_id,
-                    self.server_io.collaboration_id
+                    self.server_io.collaboration_id,
+                    namespace='/tasks'
                 )
 
             self.log.info(
@@ -508,33 +526,24 @@ class Node(object):
             new tasks.
         """
 
-        self.socketIO = SocketIO(
-            self.server_io.host,
-            port=self.server_io.port,
+        self.socketIO = SocketIO(request_timeout=60)
+        self.socketIO.connect(
+            url=f'{self.server_io.host}:{self.server_io.port}',
             headers=self.server_io.headers,
-            wait_for_connection=True
+            namespaces=['/tasks']
         )
 
-        # define() returns the instantiated action_handler
-        self.socket_tasks = self.socketIO.define(NodeTaskNamespace, '/tasks')
-        self.socket_tasks.set_node_worker(self)
+        # Connect to the /tasks namespace
+        self.socketIO.register_namespace(NodeTaskNamespace('/tasks'))
+        NodeTaskNamespace.node_worker_ref = self
 
         # Log the outcome
-        if self.socketIO.connected:
-            msg = 'connected to host={host} on port={port}'
-            msg = msg.format(
-                host=self.server_io.host,
-                port=self.server_io.port
-            )
-            self.log.info(msg)
+        while not self.socketIO.connected:
+            self.log.debug('waiting for socket connection...')
+            time.sleep(1)
 
-        else:
-            msg = 'could *not* connect to {host} on port={port}'
-            msg = msg.format(
-                host=self.server_io.host,
-                port=self.server_io.port
-            )
-            self.log.critical(msg)
+        self.log.info(f'connected to host={self.server_io.host} '
+                      f'on port={self.server_io.port}')
 
     def get_task_and_add_to_queue(self, task_id):
         """Fetches (open) task with task_id from the server.
@@ -556,13 +565,14 @@ class Node(object):
 
     def run_forever(self):
         """Forever check self.queue for incoming tasks (and execute them)."""
+        kill_listener = ContainerKillListener()
         try:
             while True:
                 # blocking untill a task comes available
                 # timeout specified, else Keyboard interupts are ignored
                 self.log.info("Waiting for new tasks....")
 
-                while True:
+                while not kill_listener.kill_now:
                     try:
                         task = self.queue.get(timeout=1)
                         # if no item is returned, the Empty exception is
@@ -575,14 +585,17 @@ class Node(object):
                     except Exception as e:
                         self.log.debug(e)
 
+                if kill_listener.kill_now:
+                    raise InterruptedError
+
                 # if task comes available, attempt to execute it
                 try:
                     self.__start_task(task)
                 except Exception as e:
                     self.log.exception(e)
 
-        except KeyboardInterrupt:
-            self.log.debug("Caught a keyboard interupt, shutting down...")
+        except (KeyboardInterrupt, InterruptedError):
+            self.log.info("Vnode is interrupted, shutting down...")
             self.socketIO.disconnect()
             sys.exit()
 
@@ -592,10 +605,16 @@ def run(ctx):
     """ Start the node."""
     logging.getLogger("urllib3").setLevel(logging.WARNING)
     logging.getLogger("requests").setLevel(logging.WARNING)
-    logging.getLogger("socketIO-client").setLevel(logging.WARNING)
+    logging.getLogger("engineio.client").setLevel(logging.WARNING)
+
 
     # initialize node, connect to the server using websockets
-    node = Node(ctx)
+    try:
+        node = Node(ctx)
+    except Exception as e:
+        log = logging.getLogger(logger_name(__name__))
+        log.critical(e)
+        exit(1)
 
     # put the node to work, executing tasks that are in the que
     node.run_forever()
